
// This program manages permutations of n integers.

// Choose either zero-based (0 .. n-1) or
// one-based (1 .. n) elements as you prefer.

// Choose storage using either std::array
// or std::vector as you prefer.

#include <algorithm>
#include <array>
#include <cmath>
#include <cstdlib>
#include <initializer_list>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <limits>
#include <numeric>
#include <random>
#include <type_traits>
#include <utility>
#include <vector>
using std::cout;
using std::setw;

//////////////////////////////////

// Many people will be satisfied setting one of the following
// four flags to true, and the other three to false.
#define USE_ARR_0 true // zero-based array storage
#define USE_ARR_1 true // one-based array storage
#define USE_VEC_0 true // zero-based vector storage
#define USE_VEC_1 true // one-based vector storage

#include "permu/A1_misc.h"
#include "permu/A2_misc.h"

#if USE_ARR_0
    #include "permu/B1_arr_0.h"
    #include "permu/B2_arr_0.h"
    #include "permu/B3_arr_0.h"
    #include "permu/B4_arr_0.h"
#endif

#if USE_ARR_1
    #include "permu/C1_arr_1.h"
    #include "permu/C2_arr_1.h"
    #include "permu/C3_arr_1.h"
    #include "permu/C4_arr_1.h"
#endif

#if USE_VEC_0
    #include "permu/D1_vec_0.h"
    #include "permu/D2_vec_0.h"
    #include "permu/D3_vec_0.h"
    #include "permu/D4_vec_0.h"
#endif

#if USE_VEC_1
    #include "permu/E1_vec_1.h"
    #include "permu/E2_vec_1.h"
    #include "permu/E3_vec_1.h"
    #include "permu/E4_vec_1.h"
#endif

#include "permu/F1_convert.h"
#include "permu/F2_dist.h"
#include "permu/F3_equal.h"

//////////////////////////////////

// Some useful permutations for testing.

constexpr int demo_ord {13};

#if USE_ARR_0
    permu_arr_0<demo_ord> const demo_wa
        { 2, 1, 4, 5, 6, 3, 9, 0, 12, 7, 10, 11, 8 };
    permu_arr_0<demo_ord> const demo_wb
        { 0, 7, 4, 8, 6, 3, 2, 1, 12, 5, 10, 9, 11 };
    permu_arr_0<demo_ord> const demo_wc
        { 6, 7, 12, 2, 3, 0, 9, 1, 10, 5, 11, 4, 8 };
#endif

#if USE_ARR_1
    permu_arr_1<demo_ord> const demo_xa
        { 3, 2, 5, 6, 7, 4, 10, 1, 13, 8, 11, 12, 9 };
    permu_arr_1<demo_ord> const demo_xb
        { 1, 8, 5, 9, 7, 4, 3, 2, 13, 6, 11, 10, 12 };
    permu_arr_1<demo_ord> const demo_xc
        { 7, 8, 13, 3, 4, 1, 10, 2, 11, 6, 12, 5, 9 };
#endif

#if USE_VEC_0
    permu_vec_0 const demo_ya
        { 2, 1, 4, 5, 6, 3, 9, 0, 12, 7, 10, 11, 8 };
    permu_vec_0 const demo_yb
        { 0, 7, 4, 8, 6, 3, 2, 1, 12, 5, 10, 9, 11 };
    permu_vec_0 const demo_yc
        { 6, 7, 12, 2, 3, 0, 9, 1, 10, 5, 11, 4, 8 };
#endif

#if USE_VEC_1
    permu_vec_1 const demo_za
        { 3, 2, 5, 6, 7, 4, 10, 1, 13, 8, 11, 12, 9 };
    permu_vec_1 const demo_zb
        { 1, 8, 5, 9, 7, 4, 3, 2, 13, 6, 11, 10, 12 };
    permu_vec_1 const demo_zc
        { 7, 8, 13, 3, 4, 1, 10, 2, 11, 6, 12, 5, 9 };
#endif

#include "demo/A_ctor.h"
#include "demo/B_mult.h"
#include "demo/C_invert.h"
#include "demo/D_split.h"
#include "demo/E_laminate.h"

#include "demo/F_next_prev.h"
#include "demo/G_reverse_pos.h"
#include "demo/H_rotate_pos.h"
#include "demo/I_trans_pos.h"
#include "demo/J_shuffle.h"

#include "demo/K_patterns.h"
#include "demo/L_cycles.h"
#include "demo/M_order.h"
#include "demo/N_equal.h"
#include "demo/O_reverse_val.h"

#include "demo/P_rotate_val.h"
#include "demo/Q_trans_val.h"
#include "demo/R_distance.h"
#include "demo/S_at_seek.h"
#include "demo/T_convert.h"

#include "demo/X_general.h"
#include "demo/Y_general.h"
#include "demo/Z_general.h"

//////////////////////////////////

int main () {
//  CREATE FROM PARTIAL PERMUTATION

    cout.setf (std::ios::fixed, std::ios::floatfield);
    cout.precision (6);

//    cout << "\n filler: " << filler;
    
//    demo_Y_general ();
    demo_Z_general ();

    constexpr int order {14};

#if USE_ARR_0
    if (false) {
        using pa0o = permu_arr_0<order>;

        static pa0o const i (false);
        pa0o const a (pa0o::partial ({10, 8,6, 4}));

        cout << "\n ";
        cout << "\n i:   " << i;
        cout << "\n a:   " << a;

        std::vector<int> v {10, 8, 6, 4};
        pa0o const b (pa0o::partial (v.begin(), v.end()));
        cout << "\n b:   " << b;

        cout << "\n ";

    }

#endif
#if USE_ARR_1

    if (false) {
        cout << "\n\n *** permu_arr_1 *** ";

        for (int n {0}; n < 3; ++n) {
            static permu_arr_1<order> const i (false);
            cout << "\n\n i:   " << i;
            permu_arr_1<order> const a (true);
            cout << "\n a:   " << a;
            auto const c (a.get_cycles());
            cout << "\n c: " << c;
            auto const d (c.rebuild ());
            cout << "\n d:   ";
            view_container (cout, d);
        }
    }
#endif

    cout << "\n\n";
    return 0;
}
