
struct fail {
   fail (char const * const m) 
      { cout << "\n\n *** fail *** " << m << " *** \n\n"; }   
};

template <typename container>
std::ostream & view_container (std::ostream & o, container const & a) {
    o << "{";
    for (auto const i: a) o << " " << setw (2) << i;
    o << " }";
    return o;
}

constexpr int filler {std::numeric_limits<int>::min()};

template <typename cycles_a, typename cycles_b>
bool are_conjugate (cycles_a const & a, cycles_b const & b) {
    if (a.size() != b.size()) throw fail
        {"are_conjugate ... size mismatch"};

    std::vector<int> c;
    std::copy_if (
        a.at().begin(), a.at().end(),
        std::back_inserter (c),
        [] (int i) {return i < 0;}
    );
    std::sort (c.begin(), c.end());

    std::vector<int> d;
    std::copy_if (
        b.at().begin(), b.at().end(),
        std::back_inserter (d),
        [] (int i) {return i < 0;}
    );
    std::sort (d.begin(), d.end());

    return c == d;
}

/////////////////////

constexpr bool permu_safe {true};

template <int sz> class permu_arr_0;
template <int sz> class permu_arr_1;
class permu_vec_0;
class permu_vec_1;

template <int sz> class cycles_arr_0;
template <int sz> class cycles_arr_1;
class cycles_vec_0;
class cycles_vec_1;

namespace SJT {
    template <typename permu>
    int sign (permu const & v, int const hi, int const bot) {
        int const vs {int (v.size())};

        if (hi <  0 ) throw fail {"SJT::sign ... hi <  0 "};
        if (hi >= vs) throw fail {"SJT::sign ... hi >= vs"};

        int result {+1};

        for (int i {0}; i < vs; ++i) {
            int const vi {v.at(i)};
            if (vi >= (hi + bot)) continue;

            for (int j {i + 1}; j < vs; ++j) {
                int const vj {v.at(j)};

                if (vj == vi) throw fail {"SJT::sign ... duplicate"};
                if (vj >= hi) continue;
                if (vj < vi) result *= -1;
            }
        }
        return result;
    }

    template <typename permu>
    int seek (permu const & v, int const hi, int const bot) {
        int const vs {int (v.size())};
        for (int n {0}; n < vs; ++n)
            if (v.at(n) == (hi + bot))
                return n;
        throw fail {"SJT::seek ... was not found"};
    }

    template <typename permu>
    bool next (permu & v, int const bot) {
        int const vs {int (v.size())};
        for (int hi {vs - 1}; hi >= 0; --hi) {
            int const loc {seek (v, hi, bot)};
            int const sn {sign (v, hi, bot)};
            if (sn == +1) {
                if (loc <= 0) continue;
                int & lft {v.at(loc - 1)};
                int & mid {v.at(loc)};
                if (mid < lft) continue;
                std::swap (lft, mid);
                return true;
            }
            if (sn == -1) {
                if ((loc + 1) >= vs) continue;
                int & rgt {v.at(loc + 1)};
                int & mid {v.at(loc)};
                if (mid < rgt) continue;
                std::swap (rgt, mid);
                return true;
            }
        }
        if (vs > 1) std::swap (v.at(0), v.at(1));
        return false;
    }

    template <typename permu>
    bool is_ident (permu & v, int const bot) {
        int const vs {int (v.size())};
        for (int i {0}; i < vs; ++i) {
            if (v.at(i) != (i + bot)) return false;
        }
        return true;
    }

    template <typename permu>
    bool prev (permu & v, int const bot) {
        int const vs {int (v.size())};
        if (is_ident (v, bot)) {
            if (vs > 1) std::swap (v.at(0), v.at(1));
            return false;
        }
        for (int hi {vs - 1}; hi >= 0; --hi) {
            int const loc {seek (v, hi, bot)};
            int const sn {sign (v, hi, bot)};
            if (sn == -1) {
                if (loc <= 0) continue;
                int & lft {v.at(loc - 1)};
                int & mid {v.at(loc)};
                if (mid < lft) continue;
                std::swap (lft, mid);
                return true;
            }
            if (sn == +1) {
                if ((loc + 1) >= vs) continue;
                int & rgt {v.at(loc + 1)};
                int & mid {v.at(loc)};
                if (mid < rgt) continue;
                std::swap (rgt, mid);
                return true;
            }
        }
        throw fail {"SJT::prev ... something is wrong"};
    }
}

