
// cycle structure for a permutation, stored in std::array, based at one

template <int sz>
class cycles_arr_1 {
    friend class permu_arr_1<sz>;

    std::array<int,sz*2> guts;

    cycles_arr_1 (std::array<int,sz> const & work) {
        std::array<bool,sz> avail;
        std::fill (avail.begin(), avail.end(), true);
        int next {0};

        for (int i {0}; i < sz; ++i) {
            if (! avail.at(i)) continue;
            int counter {next};
            guts.at(next++) = 0;

            int j {i};
            do {
                guts.at(next++) = j + base;
                avail.at(j) = false;
                --guts.at(counter);
                j = work.at(j) - base;
            } while (avail.at(j));
        }

        for (; next < guts.size(); ++next)
            guts.at(next) = filler;
    }
public:
    static constexpr int base {1};

    cycles_arr_1 () = delete;
    cycles_arr_1 (cycles_arr_1 const &) = default;
    cycles_arr_1 (cycles_arr_1 &&) = default;
    cycles_arr_1 & operator= (cycles_arr_1 const &) = default;
    cycles_arr_1 & operator= (cycles_arr_1 &&) = default;
    ~cycles_arr_1 () = default;

    #if USE_ARR_0
        explicit cycles_arr_1 (cycles_arr_0<sz> const & o);
    #endif
    #if USE_VEC_1
        explicit cycles_arr_1 (cycles_vec_1 const & o);
    #endif

    std::array<int,sz> rebuild () const {
        std::array<int,sz> result;

        for (int i {0}; true; ) {
            if (i == guts.size()) break;
            if (guts.at(i) == filler) break;

            int const u {guts.at(i)};
            ++i;
            int start {guts.at(i)};

            int curr {guts.at(i) - base};
            for (int j {-1}; j > u; --j) {
                ++i;
                if (guts.at(i) == filler) break;
                result.at(curr) = guts.at(i);
                curr = guts.at(i) - base;
            }
            ++i;
            result.at(curr) = start;
        }

        return result;
    }

    // warning: sz does not equal guts.size()
    static constexpr int size () { return sz; }
    std::array<int,sz*2> const & at () const { return guts; }

    int qty () const {
        int result {0};
        for (auto const i : guts) {
            if (i == filler) break;
            if (i < 0) ++result;
        }
        return result;
    }

    void view (std::ostream & o) const {
        o << "< " << setw (2) << sz;
        o << " " << base << " cy";
        for (auto const i: guts) {
            if (i == filler) break;
            o << " " << setw (2) << i;
        }
        o << "  a1>";
    }
};

template <int sz>
std::ostream & operator<< (std::ostream & o, cycles_arr_1<sz> const & c) {
    c.view (o);
    return o;
}

