
template <int sz>
permu_arr_1<sz>::permu_arr_1 (std::initializer_list <int> i) {
    auto iter (i.begin());
    int count {0};

    while (true) {
        if (count == sz) {
            if (iter == i.end()) break;
            if (permu_safe)
                throw fail {"permu_arr_1::permu_arr_1 "
                    "(initializer_list) ... too many initializers"};
        }
        if (iter == i.end()) {
            if (count == sz) break;
            if (permu_safe)
                throw fail {"permu_arr_1::permu_arr_1 "
                    "(initializer_list) ... too few initializers"};
        }
        guts[count] = *iter;
        ++iter;
        ++count;
    }

    verify_guts ();
}

template <int sz>
template <typename input_iter>
permu_arr_1<sz>::permu_arr_1 (input_iter begin, input_iter end) {
    auto iter (begin);
    int count {0};

    while (true) {
        if (count == sz) {
            if (iter == end) break;
            if (permu_safe)
                throw fail {"permu_arr_1::permu_arr_1 "
                    "(iter, iter) ... too many initializers"};
        }
        if (iter == end) {
            if (count == sz) break;
            if (permu_safe)
                throw fail {"permu_arr_1::permu_arr_1 "
                    "(iter, iter) ... too few initializers"};
        }
        guts[count] = *iter;
        ++iter;
        ++count;
    }

    verify_guts ();
}

/////////////////

template <int sz>
permu_arr_1<sz> permu_arr_1<sz>::partial (std::initializer_list <int> i) {
    permu_arr_1 result (false);

    for (auto const j: i) {
        int const k {j};
        if (permu_safe) {
            if (k < base)
                throw fail {"permu_arr_1::partial "
                    "(initializer_list) ... element too small"};
            if (k > sz)
                throw fail {"permu_arr_1::partial "
                    "(initializer_list) ... element too large"};
            int & rgak {result.guts.at(k - base)};
            if (rgak < base)
                throw fail {"permu_arr_1::partial "
                    "(initializer_list) ... element repeated"};
            rgak = filler;
        } else {
            result.guts.at(k - base) = filler;
        }
    }

    int n {0};
    for (auto const j: i) {
        while (result.guts.at(n) != filler) ++n;
        result.guts.at(n) = j;
    }

    result.verify_guts ();
    return result;
}

template <int sz>
template <typename forward_iter>
permu_arr_1<sz> permu_arr_1<sz>::partial (forward_iter begin, forward_iter end) {
    permu_arr_1 result (false);

    for (auto j (begin); j != end; ++j) {
        int const k {*j};
        if (permu_safe) {
            if (k < base) throw fail {"permu_arr_1::partial "
                "(iter, iter) ... element too small"};
            if (k > sz) throw fail {"permu_arr_1::partial "
                "(iter, iter) ... element too large"};
            int & rgak {result.guts.at(k - base)};
            if (rgak < base) throw fail {"permu_arr_1::partial "
                "(iter, iter) ... element repeated"};
            rgak = filler;
        } else {
            result.guts.at(k - base) = filler;
        }
    }

    int n {0};
    for (auto j (begin); j != end; ++j) {
        while (result.guts.at(n) != filler) ++n;
        result.guts.at(n) = *j;
    }

    result.verify_guts ();
    return result;
}

/////////////////

template <int sz>
void permu_arr_1<sz>::reverse_pos (int const lo, int const hi) {
    if (permu_safe)
        if (lo < 1) throw fail
            {"permu_arr_1::reverse_pos ... lo < 0"};
    if (hi == 0) {
        std::reverse (guts.begin() + lo - 1, guts.end());
        return;
    }
    if (permu_safe) {
        if (hi < lo) throw fail
            {"permu_arr_1::reverse_pos ... hi < lo"};
        if (hi > (guts.size() + 1)) throw fail
            {"permu_arr_1::reverse_pos ... hi > size + 1"};
    }
    std::reverse (guts.begin() + lo - 1, guts.begin() + hi - 1);
}

template <int sz>
void permu_arr_1<sz>::rotate_pos (
    int const lo, int const md, int const hi
) {
    if (permu_safe) {
        if (lo < 1) throw fail
            {"permu_arr_1::rotate_pos ... lo < 1"};
        if (md < lo) throw fail
            {"permu_arr_1::rotate_pos ... md < lo"};
        if (hi < md) throw fail
            {"permu_arr_1::rotate_pos ... hi < md"};
        if (hi > (guts.size() + 1)) throw fail
            {"permu_arr_1::rotate_pos ... hi > size + 1"};
    }
    auto const gb1 (guts.begin() - 1);
    std::rotate (gb1 + lo, gb1 + md, gb1 + hi);
}

template <int sz>
void permu_arr_1<sz>::trans_pos (int const i, int const j, int const k) {
    if (permu_safe) {
        constexpr int t {sz + 1};

        if (i < 1) throw fail {"permu_arr_1::trans_pos ... i < 1"};
        if (j < 1) throw fail {"permu_arr_1::trans_pos ... j < 1"};
        if (k < 1) throw fail {"permu_arr_1::trans_pos ... k < 1"};

        int const ik {i + k};
        int const jk {j + k};

        if (ik > t) throw fail {"permu_arr_1::trans_pos ... i + k > sz + 1"};
        if (jk > t) throw fail {"permu_arr_1::trans_pos ... j + k > sz + 1"};

        if (i < j) {
            if (ik > j) throw fail
                {"permu_arr_1::trans_pos ... i + k > j"};
        } else if (j < i) {
            if (jk > i) throw fail
                {"permu_arr_1::trans_pos ... j + k > i"};
        } else {
            return;
        }
    } else {
        if (i == j) return;
    }

    auto const gb1 (guts.begin() - 1);
    std::swap_ranges (gb1 + i, gb1 + i + k, gb1 + j);
    verify_guts ();
}

/////////////////

template <int sz>
void permu_arr_1<sz>::reverse_val (int const lo, int const hi) {
    // hi is one beyond

    if (permu_safe)
        if (lo < 1) throw fail
            {"permu_arr_1::reverse_val ... lo < 1"};

    if (hi == 0) {
        value::reverse (guts.begin(), guts.end(), lo, sz + 1);
        verify_guts ();
        return;
    }

    if (permu_safe) {
        if (hi < lo) throw fail
            {"permu_arr_1::reverse_val ... hi < lo"};
        if (hi > (guts.size() + 1)) throw fail
            {"permu_arr_1::reverse_val ... hi > size + 1"};
    }
    value::reverse (guts.begin(), guts.end(), lo, hi);
    verify_guts ();
}

template <int sz>
void permu_arr_1<sz>::rotate_val (
    int const lo, int const md, int const hi
) {
    if (permu_safe) {
        if (lo < 1) throw fail
            {"permu_arr_1::rotate_val ... lo < 1"};
        if (md < lo) throw fail
            {"permu_arr_1::rotate_val ... md < lo"};
        if (hi < md) throw fail
            {"permu_arr_1::rotate_val ... hi < md"};
        if (hi > (guts.size() + 1)) throw fail
            {"permu_arr_1::rotate_val ... hi > size + 1"};
    }

    value::rotate (guts.begin(), guts.end(), lo, md, hi);
    verify_guts ();
}

template <int sz>
void permu_arr_1<sz>::trans_val (int const i, int const j, int const k) {
    if (permu_safe) {
        if (i < 1) throw fail {"permu_arr_1::trans_val ... i < 1"};
        if (j < 1) throw fail {"permu_arr_1::trans_val ... j < 1"};
        if (k < 0) throw fail {"permu_arr_1::trans_val ... k < 0"};

        int const ik {i + k};
        int const jk {j + k};

        if (ik > (sz + 1)) throw fail {"permu_arr_1::trans_val ... i + k > sz + 1"};
        if (jk > (sz + 1)) throw fail {"permu_arr_1::trans_val ... j + k > sz + 1"};

        if (i < j) {
            if (ik > j) throw fail
                {"permu_arr_1::trans_val ... i + k > j"};
        } else if (j < i) {
            if (jk > i) throw fail
                {"permu_arr_1::trans_val ... j + k > i"};
        } else {
            return;
        }
    } else {
        if (i == j) return;
    }
    
    value::transpose (guts.begin(), guts.end(), i, j, k);
    verify_guts ();
}


