
template <int sz>
template <int tz>
permu_arr_1<sz+tz> permu_arr_1<sz>::direct_sum (permu_arr_1<tz> const & x) const {
    permu_arr_1<sz+tz> result;
    for (int i {0}; i < sz; ++i)
        result.guts.at(i) = guts.at(i);
    for (int i {0}; i < tz; ++i)
        result.guts.at(i+sz) = x.guts.at(i) + sz;
    result.verify_guts ();
    return result;
}

template <int sz>
template <int tz>
permu_arr_1<sz+tz> permu_arr_1<sz>::skew_sum (permu_arr_1<tz> const & x) const {
    permu_arr_1<sz+tz> result;
    for (int i {0}; i < sz; ++i)
        result.guts.at(i) = guts.at(i) + tz;
    for (int i {0}; i < tz; ++i)
        result.guts.at(i+sz) = x.guts.at(i);
    result.verify_guts ();
    return result;
}

template <int sz>
template <int tz>
permu_arr_1<tz> permu_arr_1<sz>::front_split () const {
    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_arr_1::front_split ... tz < 0 "};
        if (tz > sz) throw fail {"permu_arr_1::front_split ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz > 0) front_max =
            * std::max_element (guts.begin(), guts.begin() + tz);

        int back_min {sz + 1};
        if (tz < sz) back_min =
            * std::min_element (guts.begin() + tz, guts.end());

        if (front_max < back_min) {
            std::array<int, tz> front_permu;
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i);
            return permu_arr_1<tz> {front_permu};
        }
    }

    {
        int front_min {sz + 1};
        if (tz > 0) front_min =
            * std::min_element (guts.begin(), guts.begin() + tz);

        int back_max {0};
        if (tz < sz) back_max =
            * std::max_element (guts.begin() + tz, guts.end());

        if (back_max < front_min) {
            std::array<int, tz> front_permu;
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i) - back_max;
            return permu_arr_1<tz> {front_permu};
        }
    }

    throw fail {"permu_arr_1::front_split ... not splittable"};
}

template <int sz>
template <int tz>
permu_arr_1<tz> permu_arr_1<sz>::back_split () const {
    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_arr_1::back_split ... tz < 0 "};
        if (tz > sz) throw fail {"permu_arr_1::back_split ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz < sz) front_max =
            * std::max_element (guts.begin(), guts.begin() + (sz - tz));

        int back_min {sz + 1};
        if (tz > 0) back_min =
            * std::min_element (guts.begin() + (sz - tz), guts.end());

        if (front_max < back_min) {
            std::array<int, tz> back_permu;
            for (int i {0}; i < tz; ++i)
                back_permu.at(i) = guts.at(i + sz - tz) - front_max;
            return permu_arr_1<tz> {back_permu};
        }
    }

    {
        int front_min {sz + 1};
        if (tz < sz) front_min =
            * std::min_element (guts.begin(), guts.begin() + (sz - tz));

        int back_max {0};
        if (tz > 0) back_max =
            * std::max_element (guts.begin() + (sz - tz), guts.end());

        if (back_max < front_min) {
            std::array<int, tz> back_permu;
            for (int i {0}; i < tz; ++i)
                back_permu.at(i) = guts.at(i + sz - tz);
            return permu_arr_1<tz> {back_permu};
        }
    }

    throw fail {"permu_arr_1::back_split ... not splittable"};
}

template <int sz>
template <int tz>
std::pair <permu_arr_1<tz>, permu_arr_1<sz-tz>>
permu_arr_1<sz>::front_split_both () const {
    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_arr_1::front_split_both ... tz < 0 "};
        if (tz > sz) throw fail {"permu_arr_1::front_split_both ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz > 0) front_max =
            * std::max_element (guts.begin(), guts.begin() + tz);

        int back_min {sz + 1};
        if (tz < sz) back_min =
            * std::min_element (guts.begin() + tz, guts.end());

        if (front_max < back_min) {
            std::array<int, tz> front_permu;
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i);

            std::array<int, sz-tz> back_permu;
            for (int i {0}; i < (sz - tz); ++i)
                back_permu.at(i) = guts.at(i + tz) - front_max;

            return std::make_pair (
                permu_arr_1<tz> {front_permu},
                permu_arr_1<sz-tz> {back_permu}
            );
        }
    }

    {
        int front_min {sz + 1};
        if (tz > 0) front_min =
            * std::min_element (guts.begin(), guts.begin() + tz);

        int back_max {0};
        if (tz < sz) back_max =
            * std::max_element (guts.begin() + tz, guts.end());

        if (back_max < front_min) {
            std::array<int, tz> front_permu;
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i) - back_max;

            std::array<int, sz-tz> back_permu;
            for (int i {0}; i < (sz - tz); ++i)
                back_permu.at(i) = guts.at(i + tz);

            return std::make_pair (
                permu_arr_1<tz> {front_permu},
                permu_arr_1<sz-tz> {back_permu}
            );
        }
    }

    throw fail {"permu_arr_1::front_split_both ... not splittable"};
}

template <int sz>
template <int tz>
std::pair <permu_arr_1<sz-tz>, permu_arr_1<tz>>
permu_arr_1<sz>::back_split_both () const {
    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_arr_1::back_split_both ... tz < 0 "};
        if (tz > sz) throw fail {"permu_arr_1::back_split_both ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz < sz) front_max =
            * std::max_element (guts.begin(), guts.begin() + (sz - tz));

        int back_min {sz + 1};
        if (tz > 0) back_min =
            * std::min_element (guts.begin() + (sz - tz), guts.end());

        if (front_max < back_min) {
            std::array<int, tz> back_permu;
            for (int i {0}; i < tz; ++i)
                back_permu.at(i) = guts.at(i + sz - tz) - front_max;

            std::array<int, sz-tz> front_permu;
            for (int i {0}; i < (sz - tz); ++i)
                front_permu.at(i) = guts.at(i);

            return std::make_pair (
                permu_arr_1<sz-tz> {front_permu},
                permu_arr_1<tz> {back_permu}
            );
        }
    }

    {
        int front_min {sz + 1};
        if (tz < sz) front_min =
            * std::min_element (guts.begin(), guts.begin() + (sz - tz));

        int back_max {0};
        if (tz > 0) back_max =
            * std::max_element (guts.begin() + (sz - tz), guts.end());

        if (back_max < front_min) {
            std::array<int, tz> back_permu;
            for (int i {0}; i < tz; ++i)
                back_permu.at(i) = guts.at(i + sz - tz);

            std::array<int, sz-tz> front_permu;
            for (int i {0}; i < (sz - tz); ++i)
                front_permu.at(i) = guts.at(i) - back_max;

            return std::make_pair (
                permu_arr_1<sz-tz> {front_permu},
                permu_arr_1<tz> {back_permu}
            );
        }
    }

    throw fail {"permu_arr_1::back_split_both ... not splittable"};
}

template <int sz>
bool permu_arr_1<sz>::front_splittable (int const tz) const {
    if (tz == 0) return true;
    if (tz == sz) return true;

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_arr_1::front_splittable ... tz < 0 "};
        if (tz > sz) throw fail {"permu_arr_1::front_splittable ... tz > sz"};
    }

    int const front_max {* std::max_element (guts.begin(), guts.begin() + tz)};
    int const back_min  {* std::min_element (guts.begin() + tz, guts.end())};
    if (front_max < back_min) return true;

    int const front_min {* std::min_element (guts.begin(), guts.begin() + tz)};
    int const back_max  {* std::max_element (guts.begin() + tz, guts.end())};
    if (back_max < front_min) return true;

    return false;
}

template <int sz>
bool permu_arr_1<sz>::back_splittable (int const tz) const {
    if (tz == 0) return true;
    if (tz == sz) return true;

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_arr_1::back_splittable ... tz < 0 "};
        if (tz > sz) throw fail {"permu_arr_1::back_splittable ... tz > sz"};
    }

    int const front_max {* std::max_element (guts.begin(), guts.begin() + (sz - tz))};
    int const back_min  {* std::min_element (guts.begin() + (sz - tz), guts.end())};
    if (front_max < back_min) return true;

    int const front_min {* std::min_element (guts.begin(), guts.begin() + (sz - tz))};
    int const back_max  {* std::max_element (guts.begin() + (sz - tz), guts.end())};
    if (back_max < front_min) return true;

    return false;
}

template <int sz>
template <int tz>
permu_arr_1<sz*tz> permu_arr_1<sz>::laminate (permu_arr_1<tz> const & x) const {
    std::array<int, sz*tz> nova;

    for (int j {0}; j < tz; ++j) {
        int const offset {sz * (x.guts.at(j) - 1)};
        int const js {j * sz};

        for (int i {0}; i < sz; ++i) {
            nova.at (js + i) = offset + this->guts.at (i);
        }
    }


    return permu_arr_1<sz*tz> {nova};
}
