
// cycle structure for a permutation, stored in std::vector, based at zero

class cycles_vec_0 {
    friend class permu_vec_0;

    int sz;
    std::vector<int> guts;

    cycles_vec_0 (std::vector<int> const & work) :
        sz {int (work.size())}
    {
        std::vector<bool> avail (sz, true);

        for (int i {0}; i < sz; ++i) {
            if (! avail.at(i)) continue;
            int counter {int (guts.size())};
            guts.push_back (0);

            int j {i};
            do {
                guts.push_back(j + base);
                avail.at(j) = false;
                --guts.at(counter);
                j = work.at(j) - base;
            } while (avail.at(j));
        }
    }
public:
    static constexpr int base {0};

    cycles_vec_0 () = delete;
    cycles_vec_0 (cycles_vec_0 const &) = default;
    cycles_vec_0 (cycles_vec_0 &&) = default;
    cycles_vec_0 & operator= (cycles_vec_0 const &) = default;
    cycles_vec_0 & operator= (cycles_vec_0 &&) = default;
    ~cycles_vec_0 () = default;

    #if USE_VEC_1
        explicit cycles_vec_0 (cycles_vec_1 const & o);
    #endif
    #if USE_ARR_0
        template <int tz>
            explicit cycles_vec_0 (cycles_arr_0<tz> const & o);
    #endif

    std::vector<int> rebuild () const {
        std::vector<int> result (sz);

        for (int i {0}; true; ) {
            if (i == guts.size()) break;
            int const u {guts.at(i)};
            ++i;
            int start {guts.at(i)};

            int curr {guts.at(i) - base};
            for (int j {-1}; j > u; --j) {
                ++i;
                result.at(curr) = guts.at(i);
                curr = guts.at(i) - base;
            }
            ++i;
            result.at(curr) = start;
        }

        return result;
    }

    // warning: sz generally does not equal guts.size()
    int const size () const { return sz; }
    std::vector<int> const & at () const { return guts; }

    int qty () const {
        int result {0};
        for (auto const i : guts) {
            if (i < 0) ++result;
        }
        return result;
    }

    void view (std::ostream & o) const {
        o << "< " << setw (2) << sz;
        o << " " << base << " cy";
        for (auto const i: guts) o << " " << setw (2) << i;
        o << "  v0>";
    }
};

std::ostream & operator<< (std::ostream & o, cycles_vec_0 const & c) {
    c.view (o);
    return o;
}

