
permu_vec_0 permu_vec_0::partial (
    int const sz, std::initializer_list <int> i
) {
    permu_vec_0 result (sz, false);

    for (auto const j: i) {
        int const k {j};
        if (permu_safe) {
            if (k < 0)
                throw fail {"permu_vec_0::partial "
                    "(initializer_list) ... element too small"};
            if (k >= sz)
                throw fail {"permu_vec_0::partial "
                    "(initializer_list) ... element too large"};
            int & rgak {result.guts.at(k - base)};
            if (rgak < base)
                throw fail {"permu_vec_0::partial "
                    "(initializer_list) ... element repeated"};
            rgak = filler;
        } else {
            result.guts.at(k - base) = filler;
        }
    }

    int n {0};
    for (auto const j: i) {
        while (result.guts.at(n) != filler) ++n;
        result.guts.at(n) = j;
    }

    result.verify_guts ();
    return result;
}

template <typename forward_iter>
permu_vec_0 permu_vec_0::partial (
    int const sz, forward_iter begin, forward_iter end)
{
    permu_vec_0 result (sz, false);

    for (auto j (begin); j != end; ++j) {
        int const k {*j};
        if (permu_safe) {
            if (k < 0) throw fail {"permu_vec_0::partial "
                "(iter, iter) ... element too small"};
            if (k >= sz) throw fail {"permu_vec_0::partial "
                "(iter, iter) ... element too large"};
            int & rgak {result.guts.at(k - base)};
            if (rgak < base) throw fail {"permu_vec_0::partial "
                "(iter, iter) ... element repeated"};
            rgak = filler;
        } else {
            result.guts.at(k - base) = filler;
        }
    }

    int n {0};
    for (auto j (begin); j != end; ++j) {
        while (result.guts.at(n) != filler) ++n;
        result.guts.at(n) = *j;
    }

    result.verify_guts ();
    return result;
}

/////////////////

void permu_vec_0::reverse_pos (int const lo, int const hi) {
    if (permu_safe)
        if (lo < 0) throw fail
            {"permu_vec_0::reverse_pos ... lo < 0"};
    if (hi == -1) {
        std::reverse (guts.begin() + lo, guts.end());
        return;
    }
    if (permu_safe) {
        if (hi < lo) throw fail
            {"permu_vec_0::reverse_pos ... hi < lo"};
        if (hi > guts.size()) throw fail
            {"permu_vec_0::reverse_pos ... hi > size"};
    }
    std::reverse (guts.begin() + lo, guts.begin() + hi);
}

void permu_vec_0::rotate_pos (
    int const lo, int const md, int const hi
) {
    if (permu_safe) {
        if (lo < 0) throw fail
            {"permu_vec_0::rotate_pos ... lo < 0"};
        if (md < lo) throw fail
            {"permu_vec_0::rotate_pos ... md < lo"};
        if (hi < md) throw fail
            {"permu_vec_0::rotate_pos ... hi < md"};
        if (hi > guts.size()) throw fail
            {"permu_vec_0::rotate_pos ... hi > size"};
    }
    auto const gb0 (guts.begin() - 0);
    std::rotate (gb0 + lo, gb0 + md, gb0 + hi);
}

void permu_vec_0::trans_pos (int const i, int const j, int const k) {
    if (permu_safe) {
        int const sz {int (guts.size ())};

        if (i < 0) throw fail {"permu_vec_0::trans_pos ... i < 0"};
        if (j < 0) throw fail {"permu_vec_0::trans_pos ... j < 0"};
        if (k < 0) throw fail {"permu_vec_0::trans_pos ... k < 0"};

        int const ik {i + k};
        int const jk {j + k};

        if (ik > sz) throw fail {"permu_vec_0::trans_pos ... i + k > sz"};
        if (jk > sz) throw fail {"permu_vec_0::trans_pos ... j + k > sz"};

        if (i < j) {
            if (ik > j) throw fail
                {"permu_vec_0::trans_pos ... i + k > j"};
        } else if (j < i) {
            if (jk > i) throw fail
                {"permu_vec_0::trans_pos ... j + k > i"};
        } else {
            return;
        }
    } else {
        if (i == j) return;
    }

    auto const gb0 (guts.begin() - 0);
    std::swap_ranges (gb0 + i, gb0 + i + k, gb0 + j);
    verify_guts ();
}

/////////////////

void permu_vec_0::reverse_val (int const lo, int const hi) {
    // hi is one beyond

    int const sz {int (guts.size())};

    if (permu_safe)
        if (lo < 0) throw fail
            {"permu_vec_0::reverse_val ... lo < 0"};

    if (hi == -1) {
        value::reverse (guts.begin(), guts.end(), lo, sz);
        verify_guts ();
        return;
    }

    if (permu_safe) {
        if (hi < lo) throw fail
            {"permu_vec_0::reverse_val ... hi < lo"};
        if (hi > guts.size()) throw fail
            {"permu_vec_0::reverse_val ... hi > size"};
    }
    value::reverse (guts.begin(), guts.end(), lo, hi);
    verify_guts ();
}

void permu_vec_0::rotate_val (
    int const lo, int const md, int const hi
) {
    if (permu_safe) {
        if (lo < 0) throw fail
            {"permu_vec_0::rotate_val ... lo < 0"};
        if (md < lo) throw fail
            {"permu_vec_0::rotate_val ... md < lo"};
        if (hi < md) throw fail
            {"permu_vec_0::rotate_val ... hi < md"};
        if (hi > guts.size()) throw fail
            {"permu_vec_0::rotate_val ... hi > size"};
    }

    value::rotate (guts.begin(), guts.end(), lo, md, hi);
    verify_guts ();
}

void permu_vec_0::trans_val (int const i, int const j, int const k) {
    int const sz {int (guts.size())};

    if (permu_safe) {
        if (i < 0) throw fail {"permu_vec_0::trans_val ... i < 0"};
        if (j < 0) throw fail {"permu_vec_0::trans_val ... j < 0"};
        if (k < 0) throw fail {"permu_vec_0::trans_val ... k < 0"};

        int const ik {i + k};
        int const jk {j + k};

        if (ik > sz) throw fail {"permu_vec_0::trans_val ... i + k > sz"};
        if (jk > sz) throw fail {"permu_vec_0::trans_val ... j + k > sz"};

        if (i < j) {
            if (ik > j) throw fail
                {"permu_vec_0::trans_val ... i + k > j"};
        } else if (j < i) {
            if (jk > i) throw fail
                {"permu_vec_0::trans_val ... j + k > i"};
        } else {
            return;
        }
    } else {
        if (i == j) return;
    }
    
    value::transpose (guts.begin(), guts.end(), i, j, k);
    verify_guts ();
}


