
permu_vec_1 permu_vec_1::direct_sum (permu_vec_1 const & x) const {
    int const sz {int (guts.size())};
    int const tz {int (x.guts.size())};

    permu_vec_1 result (sz+tz);
    for (int i {0}; i < sz; ++i)
        result.guts.at(i) = guts.at(i);
    for (int i {0}; i < tz; ++i)
        result.guts.at(i+sz) = x.guts.at(i) + sz;
    result.verify_guts ();
    return result;
}

permu_vec_1 permu_vec_1::skew_sum (permu_vec_1 const & x) const {
    int const sz {int (guts.size())};
    int const tz {int (x.guts.size())};

    permu_vec_1 result(sz+tz);
    for (int i {0}; i < sz; ++i)
        result.guts.at(i) = guts.at(i) + tz;
    for (int i {0}; i < tz; ++i)
        result.guts.at(i+sz) = x.guts.at(i);
    result.verify_guts ();
    return result;
}

permu_vec_1 permu_vec_1::front_split (int const tz) const {
    int const sz {int (guts.size())};

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_vec_1::front_split ... tz < 0 "};
        if (tz > sz) throw fail {"permu_vec_1::front_split ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz > 0) front_max =
            * std::max_element (guts.begin(), guts.begin() + tz);

        int back_min {sz + 1};
        if (tz < sz) back_min =
            * std::min_element (guts.begin() + tz, guts.end());

        if (front_max < back_min) {
            std::vector<int> front_permu (tz);
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i);
            return permu_vec_1 {front_permu};
        }
    }

    {
        int front_min {sz + 1};
        if (tz > 0) front_min =
            * std::min_element (guts.begin(), guts.begin() + tz);

        int back_max {0};
        if (tz < sz) back_max =
            * std::max_element (guts.begin() + tz, guts.end());

        if (back_max < front_min) {
            std::vector<int> front_permu (tz);
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i) - back_max;
            return permu_vec_1 {front_permu};
        }
    }

    throw fail {"permu_vec_1::front_split ... not splittable"};
}

permu_vec_1 permu_vec_1::back_split (int const tz) const {
    int const sz {int (guts.size())};

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_vec_1::back_split ... tz < 0 "};
        if (tz > sz) throw fail {"permu_vec_1::back_split ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz < sz) front_max =
            * std::max_element (guts.begin(), guts.begin() + (sz - tz));

        int back_min {sz + 1};
        if (tz > 0) back_min =
            * std::min_element (guts.begin() + (sz - tz), guts.end());

        if (front_max < back_min) {
            std::vector<int> nova (tz);
            for (int i {0}; i < tz; ++i)
                nova.at(i) = guts.at(i + sz - tz) - front_max;
            return permu_vec_1 {nova};
        }
    }

    {
        int front_min {sz + 1};
        if (tz < sz) front_min =
            * std::min_element (guts.begin(), guts.begin() + (sz - tz));

        int back_max {0};
        if (tz > 0) back_max =
            * std::max_element (guts.begin() + (sz - tz), guts.end());

        if (back_max < front_min) {
            std::vector<int> nova (tz);
            for (int i {0}; i < tz; ++i)
                nova.at(i) = guts.at(i + sz - tz);
            return permu_vec_1 {nova};
        }
    }

    throw fail {"permu_vec_1::back_split ... not splittable"};
}

std::pair <permu_vec_1, permu_vec_1>
permu_vec_1::front_split_both (int const tz) const {
    int const sz {int (guts.size())};

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_vec_1::front_split_both ... tz < 0 "};
        if (tz > sz) throw fail {"permu_vec_1::front_split_both ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz > 0) front_max =
            * std::max_element (guts.begin(), guts.begin() + tz);

        int back_min {sz + 1};
        if (tz < sz) back_min =
            * std::min_element (guts.begin() + tz, guts.end());

        if (front_max < back_min) {
            std::vector<int> front_permu (tz);
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i);

            std::vector<int> back_permu (sz - tz);
            for (int i {0}; i < (sz - tz); ++i)
                back_permu.at(i) = guts.at(i + tz) - front_max;

            return std::make_pair (
                permu_vec_1 {front_permu},
                permu_vec_1 {back_permu}
            );
        }
    }

    {
        int front_min {sz + 1};
        if (tz > 0) front_min =
            * std::min_element (guts.begin(), guts.begin() + tz);

        int back_max {0};
        if (tz < sz) back_max =
            * std::max_element (guts.begin() + tz, guts.end());

        if (back_max < front_min) {
            std::vector<int> front_permu (tz);
            for (int i {0}; i < tz; ++i)
                front_permu.at(i) = guts.at(i) - back_max;

            std::vector<int> back_permu (sz - tz);
            for (int i {0}; i < (sz - tz); ++i)
                back_permu.at(i) = guts.at(i + tz);

            return std::make_pair (
                permu_vec_1 {front_permu},
                permu_vec_1 {back_permu}
            );
        }
    }

    throw fail {"permu_vec_1::front_split_both ... not splittable"};
}

std::pair <permu_vec_1, permu_vec_1>
permu_vec_1::back_split_both (int const tz) const {
    int const sz {int (guts.size())};

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_vec_1::back_split_both ... tz < 0 "};
        if (tz > sz) throw fail {"permu_vec_1::back_split_both ... tz > sz"};
    }

    {
        int front_max {0};
        if (tz < sz) front_max =
            * std::max_element (guts.begin(), guts.begin() + (sz - tz));

        int back_min {sz + 1};
        if (tz > 0) back_min =
            * std::min_element (guts.begin() + (sz - tz), guts.end());

        if (front_max < back_min) {
            std::vector<int> back_permu (tz);
            for (int i {0}; i < tz; ++i)
                back_permu.at(i) = guts.at(i + sz - tz) - front_max;

            std::vector<int> front_permu (sz - tz);
            for (int i {0}; i < (sz - tz); ++i)
                front_permu.at(i) = guts.at(i);

            return std::make_pair (
                permu_vec_1 {front_permu},
                permu_vec_1 {back_permu}
            );
        }
    }

    {
        int front_min {sz + 1};
        if (tz < sz) front_min =
            * std::min_element (guts.begin(), guts.begin() + (sz - tz));

        int back_max {0};
        if (tz > 0) back_max =
            * std::max_element (guts.begin() + (sz - tz), guts.end());

        if (back_max < front_min) {
            std::vector<int> back_permu (tz);
            for (int i {0}; i < tz; ++i)
                back_permu.at(i) = guts.at(i + sz - tz);

            std::vector<int> front_permu (sz - tz);
            for (int i {0}; i < (sz - tz); ++i)
                front_permu.at(i) = guts.at(i) - back_max;

            return std::make_pair (
                permu_vec_1 {front_permu},
                permu_vec_1 {back_permu}
            );
        }
    }

    throw fail {"permu_vec_1::back_split_both ... not splittable"};
}

bool permu_vec_1::front_splittable (int const tz) const {
    int const sz {int (guts.size())};

    if (tz == 0 ) return true;
    if (tz == sz) return true;

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_vec_1::front_splittable ... tz < 0 "};
        if (tz > sz) throw fail {"permu_vec_1::front_splittable ... tz > sz"};
    }

    int const front_max {* std::max_element (guts.begin(), guts.begin() + tz)};
    int const back_min  {* std::min_element (guts.begin() + tz, guts.end())};
    if (front_max < back_min) return true;

    int const front_min {* std::min_element (guts.begin(), guts.begin() + tz)};
    int const back_max  {* std::max_element (guts.begin() + tz, guts.end())};
    if (back_max < front_min) return true;

    return false;
}

bool permu_vec_1::back_splittable (int const tz) const {
    int const sz {int (guts.size())};

    if (tz == 0 ) return true;
    if (tz == sz) return true;

    if (permu_safe) {
        if (tz < 0 ) throw fail {"permu_vec_1::back_splittable ... tz < 0 "};
        if (tz > sz) throw fail {"permu_vec_1::back_splittable ... tz > sz"};
    }

    int const front_max {* std::max_element (guts.begin(), guts.begin() + (sz - tz))};
    int const back_min  {* std::min_element (guts.begin() + (sz - tz), guts.end())};
    if (front_max < back_min) return true;

    int const front_min {* std::min_element (guts.begin(), guts.begin() + (sz - tz))};
    int const back_max  {* std::max_element (guts.begin() + (sz - tz), guts.end())};
    if (back_max < front_min) return true;

    return false;
}

permu_vec_1 permu_vec_1::laminate (permu_vec_1 const & x) const {
    int const sz {int (guts.size())};
    int const tz {int (x.guts.size())};

    std::vector<int> nova (sz * tz);

    for (int j {0}; j < tz; ++j) {
        int const offset {sz * (x.guts.at(j) - 1)};
        int const js {j * sz};

        for (int i {0}; i < sz; ++i) {
            nova.at (js + i) = offset + this->guts.at (i);
        }
    }

    return permu_vec_1 {nova};
}
